
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <limits.h>

#include "2pc.h"

#define TIMES 15

struct sockaddr_in serv_addr;
struct sockaddr* serv;
int	ssize = sizeof(serv_addr);


/*
 * get_address() converts its host argument string (either a hostname or an
 * IP address in dotted decimal format), into an internet address and returns
 * it, or FALSE on failure.
 */
unsigned long get_address(char* host)
{
        struct in_addr  *addr_ptr;
        unsigned long   inaddr;

        /*
         * First check if host string we've been passed is in
         * dotted decimal format. Failing that, we'll get the
         * IP from the system.
         */
        if ((int)(inaddr = inet_addr(host)) == INADDR_NONE) {
                struct hostent  *hp;
                struct in_addr  *addr;

                hp = gethostbyname(host);
                if (hp == NULL) {
                        return FALSE;
                }
                addr = (struct in_addr*) hp->h_addr;
                return addr->s_addr;
        }
        else {
                return inaddr;
        }
}


int main(int argc, char **argv)
{
	int	sock, res, len, fromsize, id;
	msg	m;
	struct sockaddr_in from;

	if (argc < 2) {
		fprintf(stderr, "Usage: %s coord_host\n", argv[0]);
		exit(1);
	}

	bzero(&serv_addr, sizeof(serv_addr));
	serv_addr.sin_family 		= AF_INET;
	serv_addr.sin_addr.s_addr 	= get_address(argv[1]);
	serv_addr.sin_port		= htons(SERVER_PORT);

	serv = (struct sockaddr*) &serv_addr;

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		perror("socket");
		exit(1);
	}

	/*
	 * Register with coordinator
	 */
	fprintf(stderr, "Attempting to register...\n");
	m.type = REGISTER;
	res = sendto(sock, &m, sizeof(m), 0, serv, ssize);
	if (res < 0) {
		perror("sendto");
		exit(1);
	}

	fromsize = sizeof(from);
	res = recvfrom(sock, &m, sizeof(m), 0, 
		       (struct sockaddr*)&from, &fromsize);
	if (res < 0) {
		perror("recvfrom");
		exit(1);
	}

	fprintf(stderr, "Officially registered.\n");

	for (id = 0; id < TIMES; ) {

		m.type = PREPARED;
		m.id = id;
		m.aborted = FALSE;

		if (argc == 3 && (id == 5 || id == 10))
			m.aborted = TRUE;

		res = sendto(sock, &m, sizeof(m), 0, serv, ssize);
		if (res < 0) {
			perror("sendto");
			exit(1);
		}

		fromsize = sizeof(from);
		res = recvfrom(sock, &m, sizeof(m), 0, 
			       (struct sockaddr*)&from, &fromsize);
		if (res < 0) {
			perror("recvfrom");
			exit(1);
		}

		fprintf(stderr, "%d: got back %s message\n", id,
			m.type == COMMIT ? "COMMIT" : "ABORT");

		id = m.new_id;

		if (argc == 3)
			sleep(1);
	}
}

