#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h> /* for setitimer */
#include <sys/syscall.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/uio.h>

#include "dc.h"

/*
 * Fake out dc_init to do nothing.
 */
char *dc_stack_public;

struct timeval t_start;			/* starting time of the program */

void dc_init_internal(char *command, void (*check_ptr)(), void (*recov_ptr)(),
    char ***argvp, char ***envpp)
{
    gettimeofday(&t_start, NULL);
    dc_stack_public = STACK_START - MAX_STACK_SIZE;
    /* fprintf(stderr, "dc_stack_public = 0x%x\n", dc_stack_public); */
    setlinebuf(stdout);
}

void dc_checkpoint()
{
}

int dc_select(int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
    struct timeval *timeout)
{
    return(syscall(SYS_select, nfds, readfds, writefds, exceptfds, timeout));
}

void dc_exit(int status)
{
    exit(status);
}

ssize_t readv(int d, const struct iovec *iov, int iovcnt)
{
    int status;
    if (iovcnt > 0 && (iov[0].iov_len & 0x80000000)) {
	/* this mess is a typecast to get around the const iov */
	* (size_t *) (& (iov[0].iov_len)) &= 0x7fffffff;
    }
    status = syscall(SYS_readv, d, iov, iovcnt);
    return(status);
}

/*
 * Intercept read by transforming it into readv.
 */
ssize_t read(int d, void *buf, size_t nbytes)
{
    struct iovec iov;

    iov.iov_base = buf;
    iov.iov_len = nbytes;

    return(readv(d, &iov, 1));
}


/*
 * exit.  Only used for performance measurements (to print final statistics).
 * ???
 */
#include "atexit.h"

void (*__cleanup)();

/*
 * Exit, flushing stdio buffers if necessary.
 */
void
exit(status)
	int status;
{
	register struct atexit *p;
	register int n;
	struct timeval t_end;

	for (p = __atexit; p; p = p->next)
		for (n = p->ind; --n >= 0;)
			(*p->fns[n])();

	gettimeofday(&t_end, NULL);

	/*
	 * Print final statistics
	 */
	if (freopen("/tmp/dc.finalStats.tmp", "w", stderr) != stderr) {
	    perror("freopen");
	    _exit(1);
	}

	/*
	 * Print the type of recovery protocol that was used in this run.
	 */
#ifdef DC_DISK
	fprintf(stderr, "DC_DISK\n");
#else
	fprintf(stderr, "No DC_DISK\n");
#endif

#ifdef APP_MPROTECT
	fprintf(stderr, "APP_MPROTECT\n");
#else
	fprintf(stderr, "No APP_MPROTECT\n");
#endif

#if !defined(TRACK_NA) && !defined(TRACK_ND) && !defined(REDO_LOG) && !defined(TWO_PHASE_COMMIT)
	fprintf(stderr, "BASE\n");
#else
#	error "unknown recovery protocol"
#endif

	fprintf(stderr, "duration %lf seconds\n",
	    (t_end.tv_sec - t_start.tv_sec) +
	    (t_end.tv_usec - t_start.tv_usec) / 1000000.0);

	fprintf(stderr, "numCheckpoints 0\n");

	/*
	 * Don't count the last interval towards the average.
	 */
	fprintf(stderr, "lastIntervalBytesLogged 0\n");

	fprintf(stderr, "maxBytesLogged 0\n");
	fprintf(stderr, "averageBytesLogged 0\n");
	if (__cleanup)
		(*__cleanup)();
	_exit(status);
}
